#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include "sizes.h"

#define p __attribute__((packed))
struct BootSector
{
	BYTE BS_jmpBoot[3] p;
	BYTE BS_oemName[8] p;
	WORD BPB_BytsPerSec p; // 512
	BYTE BPB_SecPerClus p; // 1,2,4,8,16,32,64 or 128
	WORD BPB_RsvdSecCnt p; // FAT12,16:1; FAT32:32
	BYTE BPB_NumFATs p;    // 2
	WORD BPB_RootEntCnt p; // Num of 32-byte entries. For FAT16:512. For FAT32:0
	WORD BPB_TotSec16 p;   // 0
	BYTE BPB_Media p;      // Same as fat[0][0]
	WORD BPB_FATSz16 p p;    // For FAT32:0, FAT16:sectperfat
	WORD BPB_SecPerTrk p;
	WORD BPB_NumHeads p;
	DWORD BPB_HiddSec p;   // Between boot sector and fat. Should be 0
	DWORD BPB_TotSec32 p;  // Size of disk in sectors
	/* The next only to FAT32 */
	DWORD BPB_FATSz32 p;   // Could of sectors in one fat
	WORD BPB_ExtFlags p;   // bits 0-3: 0+ # of active fat.
	                       //           valid if mirroring disabled.
	                       // bits 4-6: 0
	                       // bit 7: 0=runtime mirror fats
	                       //        1=only one active fat
	                       // bits 8-15: 0
	WORD BPB_FSVer p;      // uh, major.minor version number
	DWORD BPB_RootClust p; // Usually 2. First clust of root dir.
	WORD BPB_FSInfo p;     // Usually 1.
	WORD BPB_BkBootSec p;  // Usually 6. Backupped boot sector number
			  			   //            starting from the start of
						   //            reserved area
	BYTE BPB_Reserved[12] p;// With 0
	BYTE BS_DrvNum p;      //Uh, say 0x80
	BYTE BS_Reserved1 p;
	BYTE BS_BootSig p;
	DWORD BS_VolID p;
	BYTE BS_VolLab[11] p;   //11 chars, space padded
	BYTE BS_FilSysType[8] p;//Always "FAT32   "
} BootSector;

typedef struct FSInfo
{
	DWORD FSI_LeadSig p;			// 0x41615252
	BYTE FSI_Reserved1[480] p;	// With 0
	DWORD FSI_StrucSig p;			// 0x61417272
	DWORD FSI_Free_Count p;		// Set to 0 if don't know free clust count
	DWORD FSI_Nxt_Free p;			// Hint. Set to 0 if not sure...
	BYTE FSI_Reserved2[12] p;		// With 0
	DWORD FSI_TrailSig p;			// 0xAA550000
} FSInfo;

static DWORD FirstDataSector;
static DWORD DataSec;			// Count of data sectors on data region
static DWORD CountofClusters;   // Count of data clusters (starting from 2)
static DWORD LastValidCluster;  // CountofClusters+1
static DWORD BytesPerClust;

static int fd;
static int Verbose=0;

#define VERBOSE(n) if(Verbose>=n)

static void ReadSect(void *buf, DWORD ind)
{
	lseek(fd, ind*512, SEEK_SET);
	read(fd, buf, 512);
}

static void InitData(void)
{
	// The first data cluster is cluster 2.
	// The first sector of cluster 2 is:
	FirstDataSector = BootSector.BPB_RsvdSecCnt +
    	              (BootSector.BPB_NumFATs * BootSector.BPB_FATSz32);
    	              
	BytesPerClust = BootSector.BPB_SecPerClus * BootSector.BPB_BytsPerSec;
	
	//FirstSectorofCluster(N) = ((N - 2) * BPB_SecPerClus) + FirstDataSector;

	DataSec = BootSector.BPB_TotSec32 - FirstDataSector;
	CountofClusters = DataSec / BootSector.BPB_SecPerClus;
	LastValidCluster = CountofClusters + 1;
}

static DWORD ReadFAT(DWORD clus)
{
	DWORD FATOffset = clus * 4;
	DWORD ThisFATSecNum = BootSector.BPB_RsvdSecCnt +
							(FATOffset / BootSector.BPB_BytsPerSec);
	DWORD ThisFATEntOffset = FATOffset % BootSector.BPB_BytsPerSec;
	
	DWORD data;	
	
	if(clus > LastValidCluster)
	{
		fprintf(stderr, "ReadFAT: Last clust=%lu, tried %lu.\n",
			LastValidCluster, clus);
		return 0;
	}
	
	lseek(fd, ThisFATSecNum*512+ThisFATEntOffset, SEEK_SET);
	read(fd, &data, 4);
	
	return data & 0x0FFFFFFF;
	// Note: High 4 bits shall NEVER be changed.
}

static void WriteFAT(DWORD clus, DWORD newdata)
{
	DWORD FATOffset = clus * 4;
	DWORD ThisFATSecNum = BootSector.BPB_RsvdSecCnt +
							(FATOffset / BootSector.BPB_BytsPerSec);
	DWORD ThisFATEntOffset = FATOffset % BootSector.BPB_BytsPerSec;
	
	DWORD data;
	
	if(clus > LastValidCluster)
	{
		fprintf(stderr, "WriteFAT: Last clust=%lu, tried %lu.\n",
			LastValidCluster, clus);
		return;
	}
	
	lseek(fd, ThisFATSecNum*512+ThisFATEntOffset, SEEK_SET);
	read(fd, &data, 4);
	data = (data & 0xF0000000) | (newdata & 0x0FFFFFFF);
	lseek(fd, -4, SEEK_CUR);
	write(fd, &data, 4);
	// Note: High 4 bits shall NEVER be changed.
}

// For FAT32, end of chain -marker is with this rule:
//FATContent >= 0x0FFFFFF8
// Note ignoring 4 high bits again.
// Bad cluster == 0x0FFFFFF7

// FAT[0] should be 0x0FFFFFF8
// FAT[1] should be 0x0FFFFFF8
//                  0x08000000 = ClnShutBitMask = set if clean, zero if dirty
//                  0x04000000 = HrdErrBitMask  = set if bad, zero if good

//FirstRootDirClust = BPB_RootClus;

typedef struct DirEnt
{
	BYTE DIR_Name[11] p;	// Short name
	BYTE DIR_Attr p;		// 1=R, 2=H, 4=S, 8=V, 16=D, 32=A, 15=long name
	BYTE DIR_NTRes p;		// 0
	BYTE DIR_CrtTimeTenth p;
	WORD DIR_CrtTime p;
	
	WORD DIR_CrtDate p;
	WORD DIR_LstAccDate p;
	WORD DIR_FstClusHi p; // High word of the starting cluster number
	WORD DIR_WrtTime p;
	
	WORD DIR_WrtDate p;
	WORD DIR_FstClusLo p; // Low word of the starting cluster number
	DWORD DIR_FileSize p; // In bytes
} DirEnt;

static void Save(DWORD Clust1, DWORD ClustCount)
{
	DWORD d;
	while(ClustCount>1)
	{
		DWORD next = Clust1+1;
		d = ReadFAT(Clust1);
		if(d)
		{
			if(d==next)
			{
				VERBOSE(2)
					printf("\tAlready ok...\n");
			}
			else
			{
				printf("\tWhoops, a collision. Ouch...\n");
				return;
			}
		}
		WriteFAT(Clust1, next);
		
		d = ReadFAT(next);
		if(d)
		{
			if((ClustCount>2 && d!=next+1)
			|| (ClustCount==2&& d!=0x0FFFFFF8))
			{
				printf(
					"\tWhoops and ouch. Cutting chain, lost %lu bytes\n",
						ClustCount * BytesPerClust);
				goto Terminate;
			}
			VERBOSE(2)
				printf("\tAnd river flows down...\n");
		}
		Clust1 = next;
		ClustCount--;
	}
	d = ReadFAT(Clust1);
	if(d)
	{
		if(d==0x0FFFFFF8)
		{
			VERBOSE(2)
				printf("\tAlready ok...\n");
		}
		else
		{
			printf("\tWhoops, a collision. Ouch...\n");
			return;
		}
	}
Terminate:
	WriteFAT(Clust1, 0x0FFFFFF8);
}

static DWORD Clust2Sec(DWORD Clust)
{
	return (Clust-2) * BootSector.BPB_SecPerClus + FirstDataSector;
}

static void ScanDir(DWORD ClustNum, DWORD Parent)
{
	BYTE a, b = BootSector.BPB_SecPerClus;
	DWORD next, Sec;
	DirEnt e[16];
	
Jes:Sec = Clust2Sec(ClustNum);
	for(a=0; a<b; a++)
	{
		BYTE c;
		ReadSect(&e, Sec);
		
		for(c=0; c<16; c++)
		{
			DWORD Clus;
			
			if(e[c].DIR_Attr&8)continue;	// LFN or volumename
			switch(e[c].DIR_Name[0])
			{
				case 0xE5:
					continue;				// Deleted
				case 0x05:
					e[c].DIR_Name[0] = 0xE5;
					break;
				case 0x00:
					goto Done;
			}
			printf("%-11.11s", e[c].DIR_Name);
			
			Clus = e[c].DIR_FstClusLo +
			      (e[c].DIR_FstClusHi<<16);
			      
			if(Clus > LastValidCluster)
			{
				printf(" - points out of disk (cluster %lu)\n", Clus);
				continue;
			}
			          
			if(e[c].DIR_Attr&16)
			{
				printf("\\");
				if(e[c].DIR_Name[0]=='.')
				{
					DWORD Wanted;
					Wanted = (e[c].DIR_Name[1]=='.')
						?	Parent
						:	ClustNum;
					if(!Clus)Clus = BootSector.BPB_RootClust;
					if(Wanted != Clus)
						printf(" - Should point %lu, points %lu!",
							Wanted, Clus);
				}
				printf("\n");
				if(e[c].DIR_Name[0]!='.')ScanDir(Clus, ClustNum);
				// Dir
			}
			else
			{
				printf(" - %lu bytes\n", e[c].DIR_FileSize);
				Save(Clus, (e[c].DIR_FileSize + BytesPerClust - 1)
							/ BytesPerClust);
				// File
			}
		}
		Sec++;
	}
	
	next = ClustNum+1;
	if(ReadFAT(next))goto Done;
	printf("Peeking the directory...\n");
	lseek(fd, Clust2Sec(next)*512, SEEK_SET);
	read(fd, &e[0], 32);
	for(a=0; a<11; a++)
	{
		int c=0;
		b = e[0].DIR_Name[a];
		if((a<0x20&&a!=0x05)||strchr("\"*+,./:;<=>?[\\]|",a))c=1;
		if(e[0].DIR_Attr==15)c=0;
		if(e[0].DIR_Attr>0x2F)c=1;
		if(e[0].DIR_FileSize>1073741824)c=1; // giga
		if(c)break;
    }
    if(a==11)
    {
		WriteFAT(ClustNum, next);
		ClustNum = next;
		goto Jes;
	}
Done:
	next = ReadFAT(ClustNum);
	if(next)
	{
		if(next == 0x0FFFFFF8)
		{
			VERBOSE(2)printf("Directory was already correctly terminated...\n");
			return;
		}
		printf("Oh oh. Could not save this directory... This leaves an error.\n");
		if(next==ClustNum+1)
		{
			printf(
				"Hmm. It says it continues still one sector...\n"
				"Well, worth trying.\n");
			ClustNum = next;
			goto Jes;
		}
	}
	else
	{
		printf("Directory terminated\n");
		WriteFAT(ClustNum, 0x0FFFFFF8);
	}
}

int main(int argc, const char **argv)
{
	const char *pname = NULL, *A0=argv[0];
	int mode = O_RDONLY;
	char Buf[512];
	while(--argc)
	{
		const char *s = *++argv;
		if(*s=='-')
		{
			while(*++s)
				switch(*s)
				{
					case 'v': puts("
    Copyright (C) 1992,1999 Bisqwit (http://iki.fi/bisqwit/)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License , or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, write to the Free Software
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
");
						Verbose++;
						break;
					case 'f':
					{
						printf(
							"Would enable writing to disk. Are you ABSOLUTELY sure?\n"
							"There's no way to restore your CURRENT FAT on the disk if\n"
							"you answer yes here. Do you take the risk? ");
					Refgets:
						fgets(Buf, 511, stdin);
						if(Buf[0]=='Y')
						{
							mode = O_RDWR;
							printf("\n\"Alea iacta est\" -- Julius Caesar\n\n");
							sleep(1);
						}
						else if(Buf[0]=='y')
						{
							printf("Use capital 'Y' if you are really sure. So? ");
							goto Refgets;
						}
						else
							printf("Will read only.\n");
						break;
					}
					default: goto Usage;
				}
			continue;
Usage:		printf("Usage: %s [-vf] blockdevicewithfat32tofix\n", A0);
			return -1;
		}
		if(pname)goto Usage;
		pname = s;
	}
	if(!pname)goto Usage;
	fd = open(pname, mode);
	if(fd < 0)
	{
		perror(pname);
		return -1;
	}
	
	ReadSect(Buf, 0);
	memcpy(&BootSector, Buf, sizeof(BootSector));
	
	InitData();
	
	ScanDir(BootSector.BPB_RootClust, BootSector.BPB_RootClust);
	
	return 0;
}
